# Written by John Hoffman
# see LICENSE.txt for license information

# operates as an intermediary for a Storage object, taking its place.
# stashes received data in other spots,
# moving them to their genuine location as soon as convenient.

from cStringIO import StringIO
from threading import Thread, Condition, Event, Lock
from time import sleep
true = 1
false = 0

granularity = 64*1024   # amount to allocate per cycle length
debug = false           # spews debug messages
stress_test = false     # stress-test the code by disabling automatic move-to-place functionality
cycle_length = None     # sets the cycle_length for when running in a thread
                        # to stress-test the code, set it slow; 0.25 = 4X per second
                        # for normal use, cycle_length = None works fine. :-)
                        # another good stress-test is to not run the allocation routine at all,
                        # but only allocate on the fly; if you do that, I recommend you leave
                        # stress_test false.  Otherwise you'll wind up with a nightmare of
                        # data movement in endgame.


class ChunkList:
    def __init__(self):
        self.list = []
        self.listpos = 0
        self.toplist = []
        self.has = {}
        self.deleted = {}

    def isEmpty(self):
        return not (len(self.has) > len(self.deleted))

    def push(self, val):
        if self.has.has_key(val):
            if self.deleted.has_key(val):
                del self.deleted[val]
        else:
            self.has[val] = 1
            self.toplist.append(val)
               
    def pushbottom(self, val):
        if self.has.has_key(val):
            if self.deleted.has_key(val):
                del self.deleted[val]
        else:
            self.has[val] = 1
            self.list.append(val)

    def pop(self):
        if self.isEmpty():
            return None
        if self.toplist:
            val = self.toplist.pop(-1)
            del self.has[val]
            if self.deleted.has_key(val):
                del self.deleted[val]
                return self.pop()
            return val
        if not self.list:
            return None
        val = self.list[self.listpos]
        self.listpos += 1
        del self.has[val]
        if self.deleted.has_key(val):
            del self.deleted[val]
            return self.pop()
        return val
        
    def read(self):
        val = self.pop()
        if val is not None:
            self.push(val)
        return val

    def has_key(self, val):
        if self.deleted.has_key(val):
            return false
        return self.has.has_key(val)

    def delete(self, val):
        if not self.has.has_key(val):
            return
        self.deleted[val] = 1


class IntraStorage:
    def __init__(self, storage, piecesize, doneflag):
        self.storage = storage
        self.doneflag = doneflag
        self.totsize = storage.get_total_length()
        self.chunklength = min(piecesize,granularity) # if chunklength>piecesize Bad Things happen
        self.numchunks = int((self.totsize+self.chunklength-1L)/self.chunklength)
        self.emptylist = ChunkList()        # queue of empty chunks
        self.unallocated = ChunkList()      # queue of unallocated chunks
        self.elsewhere = {}                 # chunk: new_location
        self.inuse = {}                     # new_location: chunk
        self.lock = Lock()
        self.dirty = [false] * self.numchunks # is data in chunk; just an efficiency measure
        self.allocbuf = chr(0xFF) * self.chunklength
        self.buildlock = Lock()
        self.buildthread = false
        if debug:
            print 'IntraStorage started'
            self.debugchunk = 0             # just used for debugging
        self.allocdone = false

        if storage.was_allocated():
            self.allocdone = true

    # Storage emulation

    def set_readonly(self):
        self.storage.set_readonly()

    def get_total_length(self):
        return self.totsize

    def was_preexisting(self):
        return self.storage.was_preexisting()

    def was_allocated(self):
        return self.storage.was_allocated()

    def _intervals(self, pos, amount):
        r = []
        while amount > 0:
            chunk = int(pos/self.chunklength)
            place = pos - (chunk*self.chunklength)
            size = min (amount, self.chunklength-place)
            r += [(chunk, place, size)]
            pos += size
            amount -= size
        return r

    def read(self, pos, amount):
        r = StringIO()
        for chunk, place, size in self._intervals(pos, amount):
            if debug:
                self.debugchunk = chunk
            self.lock.acquire()
            if self.elsewhere.has_key(chunk):
                chunk = self.elsewhere[chunk]
            elif self.unallocated.has_key(chunk):      # reading unallocated area
                if debug:
                    print ' -- not there'
                r.write(self.allocbuf[:size])
                self.lock.release()
                continue
            pos = (chunk*self.chunklength) + place
            r.write(self.storage.read(pos, size))
            if debug:
                print 'read '+str(self.debugchunk)
                if chunk != self.debugchunk:
                    print ' -- from '+str(chunk)+' - '+str(place)+'-'+str(place+size)
                else:
                    print ' -- '+str(place)+'-'+str(place+size)
            self.lock.release()
        return r.getvalue()

    def _writedata(self, chunk, place, size, s, start):
        pos = (chunk*self.chunklength) + place
        self.storage.write(pos, s[start : start+size])
        self.dirty[chunk] = true
        if debug:
            print 'wrote '+str(self.debugchunk)
            if chunk != self.debugchunk:
                print ' -- to '+str(chunk)+' - '+str(place)+'-'+str(place+size)
            else:
                print ' -- '+str(place)+'-'+str(place+size)
        self.lock.release()

    def _writewholechunk(self, chunk, place, size, s, start):
        chunksize = min(self.chunklength, self.totsize-((chunk+1)*self.chunklength))
        self._writedata(chunk, 0, chunksize,    # update this later to use a static buffer
                        (chr(0xFF)*place) + s[start : start+size] +
                        (chr(0xFF)*(chunksize-(place+size))), 0)

    def _write(self, chunk, place, size, s, start):
        if self.doneflag.isSet():
            return  # don't try to write if things are shutting down.
        self.lock.acquire()
        if debug:
            self.debugchunk = chunk
        if chunk >= self.numchunks:
            raise ValueError, 'I/O error - bad location (too large)'
        if self.elsewhere.has_key(chunk):
            self._writedata(self.elsewhere[chunk],place,size,s,start)   # write to elsewhere
        elif not self.unallocated.has_key(chunk):   # my spot is in allocated space
            self.emptylist.delete(chunk)
            if self.inuse.has_key(chunk):           # hum.  It's busy.
                if debug:
                    print '--COLLISION--'
                self.lock.release()
                self._move(self.inuse[chunk], dontfree = true)
                self._write(chunk,place,size,s,start) # try again
            else:
                self._writedata(chunk,place,size,s,start)
        else:
            nextunallocated = self.unallocated.read()
            if chunk == nextunallocated:    # it's just off the top of what's already done
                self.unallocated.pop()      # pop that off the top
                self._writewholechunk(chunk,place,size,s,start) # automatically allocate the space
            elif not stress_test and not self.elsewhere.has_key(nextunallocated):
                empty = self.unallocated.pop()      # nothing slated to go in the top spot,
                self.elsewhere[chunk] = empty       # so, we'll stash this piece here instead
                self.inuse[empty] = chunk
                self._writewholechunk(empty,place,size,s,start)
            else:                                   # stash it elsewhere in allocated space
                empty = self._get_empty()
                if empty is None:                   # couldn't get one, had to allocate more
                    if not self.doneflag.isSet():   # or done, in which case, return without writing
                        self._write(chunk,place,size,s,start)   # so retry
                else:
                    self.elsewhere[chunk] = empty   # now, there is an elsewhere reference for chunk.
                    self.inuse[empty] = chunk       # and mark its spot as in-use
                    self._writewholechunk(empty,place,size,s,start)

    def write(self, pos, s):
        start = 0
        for chunk, place, size in self._intervals(pos, len(s)):
            self._write(chunk, place, size, s, start)
            start += size

    def close(self):
        self.storage.close()

    def isthere(self, pos, length):    # returns true if already allocated
        # assumes a piece is an even multiple of chunk -- fine so long as piece is a power of 2
        if self.storage.isthere(pos,length):
            return true
        else:
            for chunk, place, size in self._intervals(pos, length):
                self.unallocated.pushbottom(chunk)
            return false

    # IntraStorage-specific functions

    def IS_buildemptylist(self, statusfunc = lambda: None):
        graphstep = int(self.numchunks/50)
        statusfunc(fractionDone = 0.0)
        graphnext = graphstep
    
        for x in xrange(self.numchunks):
            self.unallocated.pushbottom(x)
            if x >= graphnext:
                graphnext += graphstep
                statusfunc(fractionDone = float(x+1)/(2*self.numchunks))    # only go to 50%
                             # setting the hash list in StorageWrapper will do the other 50%
    
    def _get_empty(self):   # called locked; returns unlocked and with None if it had to build
        empty = self.emptylist.pop()
        if empty is None:
            if debug:
                print 'waiting for empty'
            self.lock.release()
            while self.emptylist.isEmpty():
                self._IS_bump()         # make some free space
            return None
        return empty

    def _move(self, chunk, dontfree = false): # moves the data for chunk either to its
        self.lock.acquire()                   # real spot or another temporary spot.
        self._move_locked(chunk, dontfree)
        
    def _move_locked(self, chunk, dontfree = false):  # called when lock was already acquired
        if not self.elsewhere.has_key(chunk):   # true position
            self.lock.release()				# don't be afraid to call move if you think a
            return							# chunk's home spot is open
        if debug:
            print 'moving'
        oldspot = self.elsewhere[chunk]
        if self.unallocated.has_key(chunk): # move to another temporary location
            newspot = self._get_empty()
            if newspot is None:
                if self.doneflag.isSet():   # stopped while waiting for new chunk
                    return                  # exit without writing anything
                self._move(chunk)	        # otherwise, retry
                return
            self.elsewhere[chunk] = newspot # now, there is an elsewhere reference for chunk.
            self.inuse[newspot] = chunk
            del self.inuse[oldspot]
        else:                               # hey, my spot is allocated!
            if self.inuse.has_key(chunk):   # ...but it's not available!
                if debug:
                    print '==COLLISION=='   # this shouldn't happen except during stress tests
                conflict = self.inuse[chunk]
                self.lock.release()
                self._move(conflict, true)  # but just in case...
                self._move(chunk, dontfree) # get it out of the way, then try again
                return
            newspot = chunk
            self.emptylist.delete(chunk)
            del self.elsewhere[chunk]
            del self.inuse[oldspot]
        if debug:
            print ' -- '+str(chunk)+' from '+str(oldspot)+' to '+str(newspot)
        if self.dirty[oldspot]:             # efficiency measure; if there happens to be
            self.dirty[oldspot] = false     # no data written to the spot,
            self.dirty[newspot] = true      # no need to do disk access
            pos = oldspot * self.chunklength
            pos2 = newspot * self.chunklength
            amount = min(self.chunklength, self.totsize-pos2)
            buf = self.storage.read(pos, amount)
            self.storage.write(pos2, buf)
        if not dontfree:
            self.emptylist.pushbottom(oldspot)  # push the old spot onto the bottom of the free list
        self.lock.release()

    def _IS_alloc(self):
        self.lock.acquire()
        chunk = self.unallocated.pop()
        if chunk is None:
            if debug:
                print 'allocation complete!'
            self.allocdone = true
            self.lock.release()
            return
        pos = chunk*self.chunklength
        if not stress_test:
            if self.elsewhere.has_key(chunk):   # if there's data available for the chunk
                                                # above the one completely allocated,
                                                # just do a move over it, which will
                self._move_locked(chunk)        # automatically allocate that disk space
                return
        amount = min(self.chunklength, self.totsize-pos)
        try:
            self.storage.write(pos, self.allocbuf[:amount])
        except IOError, e:
            if self.doneflag.isSet():
                self.lock.release()
                return
            raise IOError, e
        except ValueError, e:
            if self.doneflag.isSet():
                self.lock.release()
                return
            raise ValueError, e
        if chunk != self.numchunks-1 :  # never push the end chunk on the free list
            self.emptylist.push(chunk)      # since it's probably smaller than standard
        self.lock.release()

    # for allocation, use either this:
    def IS_bump(self):  # to be called from a thread repeatedly until it returns true
        self.buildlock.acquire()
        retval = self._IS_bump()
        self.buildlock.release()
        return retval

    def _IS_bump(self):  # to be called from a thread repeatedly until it returns true
        if not self.allocdone:
            self._IS_alloc()
            return false
        self.lock.acquire()
        for chunk in self.elsewhere:    # shouldn't happen except while stress-testing
            if debug:
                print 'endgame:'
            self._move_locked(chunk)    # but just in case...
            return false                # only move a single chunk, then wait to loop again
        self.lock.release()
        if debug:
            print 'IntraStorage finished.'
        return true

    # ...or:

    def IS_build(self, final = false):  # to be called from a thread or from the idle list
        self.buildlock.acquire()
        if not final:                   # being called just before Storagewrapper calls finish()
            self.buildthread = true
        while true:
            if self.doneflag.isSet() or self._IS_bump():
                if self.doneflag.isSet():
                    self.storage.close()
                self.buildthread = false
                self.buildlock.release()
                return
            if cycle_length is not None:
                sleep(cycle_length)


# to use from StorageWrapper:
# self.storage = IntraStorage(storage, num_chunks, chunk_length, doneflag)
#
# Then either start a thread that calls self.storage.IS_build(), or create a function that
# schedules itself and calls storage.IS_bump() periodically until it returns true.  (By default
# IS_bump allocates 64K each time it's called; if you want to allocate more per scheduled
# operation, run it multiple times.)
#
# Finally, when either IS_build returns or IS_bump returns true, lock out StorageWrapper's
# get_chunk and chunk_came_in, set self.storage = self.storage.storage, and you're all set.
# You could also just leave it in place, since when everything's done, transfers will all
# fall through to Storage.
#
# I'd recommend, if the IntraStorage object is still in place when the torrent completes,
# also calling IS_build(true) right before setting the torrent as complete; this will
# block until allocation is complete, and also make sure all chunks are in the right place.